/*************************************************************************
 *
 * FILE  main.cpp
 * Copyright (c) 2005, 2006 Daniel Kahlin <daniel@kahlin.net>
 * Written by Daniel Kahlin <daniel@kahlin.net>
 * $Id: main.cpp,v 1.48 2007-01-02 21:53:43 tlr Exp $
 *
 * DESCRIPTION 
 *
 ******/

#include <iostream>
#include <string>
#include <list>
#include <vector>
#include <stdio.h>
#include "util/types.hpp"
#include "util/getopt.hpp"
#include "global.hpp"
#include "boot.hpp"
#include "transfer.hpp"
#include "membuffer.hpp"

#include "zipfile.h"

// arch dependant
#include "timer.hpp"
#include "port.hpp"

#define PROGRAM "dtvtrans"

#ifndef IS_LINUX
#define PORT_DEFAULT "0x378"
#else
#define PORT_DEFAULT "/dev/parport0"
#endif

using namespace std;

typedef int (*cmd_func_t)(Getopt&);
int cmd_boot(Getopt& opts);
int cmd_read_mem(Getopt& opts);
int cmd_write_mem(Getopt& opts);
int cmd_execute_mem(Getopt& opts);
int cmd_clear_port(Getopt& opts);
int cmd_upload_file(Getopt& opts);

int main(int argc, char *argv[])
{
    std::map<std::string,cmd_func_t> cmds;
    cmds.insert(std::make_pair<std::string,cmd_func_t>("boot",cmd_boot));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("bt",cmd_boot));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("b",cmd_boot));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("read",cmd_read_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("rd",cmd_read_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("r",cmd_read_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("write",cmd_write_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("wr",cmd_write_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("w",cmd_write_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("go",cmd_execute_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("g",cmd_execute_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("clear",cmd_clear_port));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("clr",cmd_clear_port));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("c",cmd_clear_port));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("u",cmd_upload_file));

    /* defaults */
    verbose_g = false;
    debug_g   = false;
    portbase_g = PORT_DEFAULT;
    blocksize_g = 0x0400;
    rawmode_g = false;

    /*
     * First scan for any options.
     *
     */
    Getopt::Opt opt;
    Getopt opts(argc, argv, "rb:p:vdVh");
    while (opt = opts.next_opt(), !opt.is_end()) {

	switch(opt.get_status()) {
	case Getopt::Opt::OK:
	    break;
	case Getopt::Opt::MISSING:
	    fprintf(stderr, PROGRAM ": missing parameter for option -%c\n",opt.get_opt());
	    return 1;
	case Getopt::Opt::INVALID:

	    fprintf(stderr, PROGRAM ": invalid option -%c\n",opt.get_opt());
	    return 1;
	default:
	    fprintf(stderr, PROGRAM ": error while parsing option -%c\n",opt.get_opt());
	    return 1;
	}

	switch (opt.get_opt()) {

        /* set raw format */
        case 'r':
	    rawmode_g = true;
            break;

        case 'b':
	    {
		const char* str=opt.get_arg();
		char* endptr;
		blocksize_g=std::strtoul(str,&endptr,0);
		if (endptr==str) {
		    fprintf(stdout, "invalid number (%s)\n",str);// failed
		    return 1;
		}
	    }
	    if (blocksize_g < 1 || blocksize_g > 0x4000) {
		fprintf(stdout, "blocksize out of range (=%d, allowable 1-16384)\n", blocksize_g );
		return -1;
	    }
            break;

        case 'p':
	    portbase_g=opt.get_arg();
            break;

        /* set verbose */
        case 'v':
	    verbose_g = true;
            break;

        /* set debug */
        case 'd':
	    debug_g = true;
            break;

        /* print version */
        case 'V':
            fprintf (stdout, PROGRAM " " VERSION "\n");
            return 0;

        /* print help */
        case 'h':
            fprintf (stdout,
PROGRAM " " VERSION "\n"
"Copyright (c) 2005, 2006, 2007 Daniel Kahlin\n"
"Written by Daniel Kahlin <daniel@kahlin.net>\n"
#ifdef IS_LINUX
"Linux backend written by Hannu Nuotio <hannu.nuotio@pp.inet.fi>\n"
"Copyright (c) 2006, 2007 Hannu Nuotio\n"
#endif
"\n"
"usage: " PROGRAM " [OPTION]... <cmd> [OPTION]... [<file>]\n"
"\n"
"Valid options:\n"
"    -r              use raw format for files (no load address)\n"
"    -b <size>       block size (default: 1024)\n"
"    -p <addr>       port address (default: " PORT_DEFAULT ")\n"
"    -v              be verbose\n"
"    -d              output debugging information\n"
"    -h              displays this help text\n"
"    -V              output program version\n"
"\n"
"Valid commands:\n"
"    boot/bt/b       send file to boot.prg/mlboot.prg\n"
"    read/rd/r       read mem\n"
"    write/wr/w      write mem\n"
"    go/g            execute mem (go)\n"
"    clear/clr/c     clear port (tristate)\n"
"    u/upload        upload file to TULF-tool running on DTV\n"
"\n");
            return 0;

        /* default behavior */
        default:
            break;
        }
    }

    if (opts.args_left()<1) {
	printf("too few arguments\n");
	return 1;
    }
    const char* cmd=opts.next_arg();

    int ret=1;
    std::map<std::string,cmd_func_t>::const_iterator it=cmds.find(cmd);
    if (it!=cmds.end()) {
	try {
	    ret=(it->second)(opts);
	} catch (...) {
	    return 1;
	}
    } else {
	fprintf(stderr, PROGRAM ": invalid operation selected (%s)\n",cmd);
	return 1;
    }

    return ret;
}


int cmd_read_mem(Getopt& opts)
{
    //bool have_prg=true;
    bool basic_load=false;

    if (opts.args_left()<1) {
	printf("too few arguments\n");
	return 1;
    }
    if (opts.args_left()<2) {
	basic_load=true;
    }

    const char* name;
    MemBuffer mb;
    Port pt(portbase_g);
    Transfer tf(pt,blocksize_g);

    if (!basic_load) {
	const char* range=opts.next_arg();
	name=opts.next_arg();

	mb.set_range(range);
    } else {
	name=opts.next_arg();
	uint8_t sabuf[2];
	uint8_t eabuf[2];
	uint32_t sa,ea;

	tf.read_mem(0,  0x002b, 2, sabuf );
	tf.read_mem(0,  0x002d, 2, eabuf );
	sa=sabuf[0] | (sabuf[1]<<8);
	ea=eabuf[0] | (eabuf[1]<<8);
	mb.set_start(sa);
	mb.set_length(ea-sa);
    }
	
    Timer tim;

    uint32_t len=mb.get_length();
    int blks=(len+2+253)/254;

    printf("transfering... $%06X-$%06X (%d bytes, %d %s)\n",mb.get_start(), mb.get_end(),mb.get_length(),blks,(blks==1)?"block":"blocks");
    fflush(stdout);
    tim.start();
    tf.read_mem( (mb.is_rom())?1:0,  mb.get_start(), mb.get_length(), mb.get_buffer() );
    tim.stop();

    float dur=static_cast<float>(tim.get_time_d());
    float speed_bytes=(len/dur);
    float speed_bits=(len/dur)*8.0f;
    printf ("transfered %d bytes in %f s (%f bytes/s, %f bit/s)\n",len,dur,speed_bytes,speed_bits);


    printf("writing file...\n");
    fflush(stdout);
    mb.write_file(name,!rawmode_g);

    return 0;
}

int cmd_write_mem(Getopt& opts)
{
    bool have_prg=!rawmode_g;
    bool set_endaddr=true;

    if (opts.args_left()<1) {
	printf("too few arguments\n");
	return 1;
    }
    
    const char* name;

    MemBuffer mb;
    if (opts.args_left()<2) {
	name=opts.next_arg();  
	mb.read_file(name,true);
    } else {
	const char* addrstr=opts.next_arg();
	name=opts.next_arg();  
	mb.read_file(name,have_prg);
	mb.set_start(addrstr);
    }

    uint32_t len=mb.get_length();
    int blks=(len+2+253)/254;

    Port pt(portbase_g);
    Transfer tf(pt,blocksize_g);
    Timer tim;

    printf("transfering... $%06X-$%06X (%d bytes, %d %s)\n",mb.get_start(), mb.get_end(),mb.get_length(),blks,(blks==1)?"block":"blocks");
    fflush(stdout);
    tim.start();
    tf.write_mem( (mb.is_rom())?1:0, mb.get_start(), mb.get_length(), mb.get_buffer() );

    if (set_endaddr) {
	uint32_t ea=mb.get_end();
	uint8_t abuf[2];
	abuf[0]=ea & 0xff;
	abuf[1]=(ea >> 8) & 0xff;
	tf.write_mem(0, 0x00ae, 2, abuf );
	tf.write_mem(0, 0x002d, 2, abuf );
    }

    tim.stop();
    float dur=static_cast<float>(tim.get_time_d());
    float speed_bytes=(len/dur);
    float speed_bits=(len/dur)*8.0f;
    printf ("transfered %d bytes in %f s (%f bytes/s, %f bit/s)\n",len,dur,speed_bytes,speed_bits);

    return 0;
}

int cmd_execute_mem(Getopt& opts)
{
    if (opts.args_left()<1) {
	printf("too few arguments\n");
	return 1;
    }
    const char* start=opts.next_arg();
    char* endptr;

    uint32_t startaddr=std::strtoul(start,&endptr,0);

    Port pt(portbase_g);
    Transfer tf(pt,blocksize_g);

    tf.execute_mem(startaddr);

    return 0;
}

int cmd_boot(Getopt& opts)
{
    if (opts.args_left()<1) {
	printf("too few arguments\n");
	return 1;
    }
    const char* name=opts.next_arg();

    MemBuffer mb;
    mb.read_file(name);

    uint32_t sa=mb.get_start();
    uint32_t len=mb.get_length();

    Port pt(portbase_g);
    Boot bt(pt);
    Timer tim;

    printf("sending boot... ($%04X-$%04X)\n",sa,sa+len-1);
    fflush(stdout);
    tim.start();
    bt.send_file(sa,sa+len-1, mb.get_buffer());
    tim.stop();
   
    float dur=tim.get_time_d();
    float speed_bytes=(len/dur);
    float speed_bits=(len/dur)*8.0f;
    printf ("sent %d bytes in %f s (%f bytes/s, %f bit/s)\n",len,dur,speed_bytes,speed_bits);

    return 0;
}

int cmd_clear_port(Getopt& opts)
{
    Port pt(portbase_g);
    pt.tristate();
    printf("Tri-stating port\n");
    return 0;
}


void str_toupper(char * dst, const char * src){
	uint8_t c;
	uint32_t cnt = 0;
	while((c = *src++)){
		*dst++ = toupper(c);
		if(cnt++ == 22) break;
	}
	*dst = 0;
}




void do_upload(const char* c64name, int startsys, MemBuffer &mb){
    Port pt(portbase_g);
    Transfer tf(pt,blocksize_g);
    Timer tim;
	
	printf("sending file info: len=0x%x, ",mb.get_length() );
	if(startsys)
		printf("startsys=%d, ", startsys);
	
	printf("name=%s\n", c64name);
	fflush(stdout);
    
    tf.write_fileinfo( mb.get_start(), mb.get_length(), blocksize_g, startsys, c64name );
	
    uint32_t len=mb.get_length();
    printf("transfering...");
    fflush(stdout);
    tim.start();

    tf.write_mem( 0, 0x020000, mb.get_length(), mb.get_buffer() );
    
    tim.stop();
    float dur=static_cast<float>(tim.get_time_d());
    float speed_bytes=(len/dur);
    float speed_bits=(len/dur)*8.0f;
    printf ("transfered %d bytes in %f s (%f bytes/s, %f bit/s)\n",len,dur,speed_bytes,speed_bits);
}

//return pointer to the extension part of a filename
const char * get_extension(const char * filename){
	int x;
	x = strlen(filename);

	do{
		x--;
	}while( (filename[x] != '.') && (x > 0) );

	if(x==0)
		return "foobar";
	
	return &filename[x+1];
}

//make a c64name form the PC name (strip extension and path)
void make_c64name(char * out, const char * in){
	const char * beg, * end, *p;
	char c;
	int size;
	char tmpbuf[24];
	
	
	//look for last slash
	beg = in;
	p = in;
	while((c = *p++)){
		if(c == '/'){
			beg = p;
		}
	}
	
	//look for last dot
	end = beg + strlen(beg);
	p = end;
	while(p != beg){
		if(*p == '.'){
			end = p;
			break;
		}
		p--;
	}
	
	size = end-beg;
	if(size > 23)
		size = 23;
	
	strncpy(tmpbuf, beg, size);
	
	tmpbuf[size] = 0;
	
	str_toupper(out, tmpbuf);
}

std::vector<std::string> tokenize_str(const std::string & str,
                                      const std::string & delims=", \t")
{
  using namespace std;
  // Skip delims at beginning, find start of first token
  string::size_type lastPos = str.find_first_not_of(delims, 0);
  // Find next delimiter @ end of token
  string::size_type pos     = str.find_first_of(delims, lastPos);

  // output vector
  vector<string> tokens;

  while (string::npos != pos || string::npos != lastPos)
    {
      // Found a token, add it to the vector.
      tokens.push_back(str.substr(lastPos, pos - lastPos));
      // Skip delims.  Note the "not_of". this is beginning of token
      lastPos = str.find_first_not_of(delims, pos);
      // Find next delimiter at end of token.
      pos     = str.find_first_of(delims, lastPos);
    }

  return tokens;
}

int cmd_upload_file(Getopt& opts)
{
    bool have_prg=!rawmode_g;

    if (opts.args_left()<1) {
		printf(	"to few arguments\n"
				"usage:\n"
				"dtvtrans [-r] u <filename.prg|pck|zip> [c64name [load-address]]\n"
		);
		return 1;
    }
    
	
    const char* name;
	const char* name_ext;
	const char* c64name;
    char mystring[24];
	MemBuffer mb;
    name=opts.next_arg();
	
	name_ext = get_extension(name);
	if( strcasecmp(name_ext, "zip") == 0){
		
		//read the index.txt file to lines
		
		list<string> lines;
		
		zip_initialize();
		zip_open_archive(name);
		zip_open_file("index.txt");
		
		while(1){		
			char * line = zip_buf_get_line();
			if(line == 0)
				break;
			lines.push_back (string(line));
		}
		
		//structure looks like this:
		//lastninja_dtv.dtv,THE LAST NINJA,2040,51287
		//ninja0b_dtv.dtv,$NINJA0B,45056,0
		
		list<string>::iterator it;
		for( it = lines.begin() ; it != lines.end() ; it++)
		{
			vector<string> substrs(tokenize_str(*it, ","));
			
			if( substrs.size() < 3)
				break;
			
			int startsys;
			
			if( substrs.size() < 4){
				startsys = 0;
			}else{
				startsys =	strtoul(substrs[3].c_str(),0 ,0 );
			}
			
			zip_open_file(substrs[0].c_str());
			mb.load_by_function( &zip_read_buffer );
			mb.set_start(substrs[2].c_str());
			
			if(substrs[1][0] == '$'){
				substrs[1] = substrs[1].substr(1);
			}
			
			do_upload(substrs[1].c_str(), startsys, mb);
		}
	
	}else if( strcasecmp(name_ext, "pck") == 0){
		mb.read_file(name, 0);
	
		if (opts.args_left() == 0) {
			make_c64name(mystring, name);
			c64name = mystring;
		} else {
			str_toupper(mystring, opts.next_arg());
			c64name = mystring;
		}
		
		if(opts.args_left() > 0){
			const char * addrstr = opts.next_arg();
			mb.set_start(addrstr);
		}else{
			printf("using default load-address of $0801\n");
			mb.set_start("0x0801");
		}
	
		do_upload(c64name, 0, mb);
	}else{ //prg
		
		mb.read_file(name, have_prg);
	
		if (opts.args_left() == 0) {
			make_c64name(mystring, name);
			c64name = mystring;
		} else {
			str_toupper(mystring, opts.next_arg());
			c64name = mystring;
		}
		
		if(opts.args_left() > 0){
			const char * addrstr = opts.next_arg();
			mb.set_start(addrstr);
		}else if(!have_prg){
			printf("using default load-address of $0801\n");
			mb.set_start("0x0801");
		}
	
		printf("packing..");
		fflush(stdout);
		
		int unpacked_length = mb.get_length();
		mb.pack(1);
		printf("encoded %d bytes into %d bytes\n", unpacked_length, mb.get_length());
		
		do_upload(c64name, 0, mb);
	}
	return 0;
}



/* eof */
