/*************************************************************************
 *
 * FILE  membuffer.cpp
 * Copyright (c) 2005, 2006 Daniel Kahlin <daniel@kahlin.net>
 * Written by Daniel Kahlin <daniel@kahlin.net>
 * $Id: membuffer.cpp,v 1.6 2006-11-29 20:10:08 tlr Exp $
 *
 * DESCRIPTION 
 *
 ******/

#include <iostream>
#include <fstream>
#include <string>
#include "util/types.hpp"
#include "membuffer.hpp"

void Address::set_start(const char* str)
{
    char* endptr;

    rom_=false;
    if (*str=='r' || *str=='R') {
	str++;
	rom_=true;
    }
    uint32_t start=std::strtoul(str,&endptr,0);
    start &=0xffffff;
    if (endptr==str || *endptr!=0) {
	// failed
	return;
    }
    start_ = start;
}


void AddressRange::set_range(const char* str)
{
    char* endptr;

    if (*str=='r' || *str=='R') {
	str++;
	rom_=true;
    }
    uint32_t start=std::strtoul(str,&endptr,0);
    start &=0xffffff;
    if (endptr==str) {
	// failed
	return;
    }
    str=endptr;
    char sep=*str++;
    if ( !(sep == '-' || sep == ',') ) {
	// failed
	return;
    }
    uint32_t end=std::strtoul(str,&endptr,0);
    end &=0xffffff;
    if (endptr==str) {
	// failed
	return;
    }
    /* parsed, set. */

    if (sep=='-') {
	start_=start;
	length_=end-start;
    }
    if (sep==',') {
	start_=start;
	length_=end;
    }
}

void MemBuffer::read_file(const char* name, bool use_prg)
{
    std::ifstream is;
    is.open(name, std::ios::binary);
    is.seekg(0, std::ios::end);
    uint32_t len=is.tellg();
    is.seekg(0, std::ios::beg);

    uint32_t sa;
    
    if (use_prg) {
	char c;
	is.get(c);
	sa=static_cast<uint8_t>(c);
	is.get(c);
	sa|=static_cast<uint8_t>(c)<<8;
	len-=2;
    }

    delete buffer_;
    buffer_=new uint8_t[len];

    for (unsigned int i=0; i < len; ++i) {
	char c;
	is.get(c);
	buffer_[i]=static_cast<uint8_t>(c);
    }

    length_=len;
    start_=sa;

    is.close();
}

void MemBuffer::load_by_function(int (*fct)(char * buf, int maxbytes) )
{
    delete buffer_;
    buffer_=new uint8_t[0x200000];//won't be any larger than the dtv's flash....
	
	length_ = fct((char*)buffer_, 0x200000);

}

void MemBuffer::write_file(const char* name, bool use_prg) const
{
    std::ofstream os(name, std::ios::binary);

    if (use_prg) {
	uint32_t sa = get_start();
	os.put( sa & 0xff );
	os.put( (sa >> 8) & 0xff );
    }

    for (unsigned int i=0; i < get_length(); ++i)
	os.put(buffer_[i]);

    os.close();
}





/*
 * File Format
 * -----------
 * The files are compressed using a simple equal-sequence compression algorithm
 *
 * Every chunk of data starts with a code-byte.
 * If the code is $01-$7f, that number of bytes follow.
 * If the code is $80-$ff, then len=code & $7f, and a second byte follows
 * this copies len bytes of data from the current destination-$0100+secondbyte.
 * If the code is $00, then we are done.
 *
 * To just store the data without any compression, put an $7f byte before every
 * 127 bytes, and terminate with a $00 byte.
 *
 */
/**************************************************************************
 *
 * SECTION  packer
 *
 ******/
#define PK_WINDOW (0x100)
#define PK_RUNLEN (0x7f)
#define PK_MAXMATCH (0x7f)

void MemBuffer::pack(bool do_pack){
	uint32_t len;
	//calculate worst case lenght of packed buffer
	len = ((length_-1)/0x7f + 1) * 0x80 + 1;
	
	//allocate that
	delete buffer_pck_;
    buffer_pck_=new uint8_t[len];

	//reset length
	pck_len_ = 0;
	
	if(do_pack){
		pk_pack_();	
	}else{
		pk_store_();	
	}
	
	delete buffer_;
	buffer_ = buffer_pck_;
	
	length_ = pck_len_;
	
}


void MemBuffer::pk_put_(uint8_t data)
{
    buffer_pck_[pck_len_] = data;
    pck_len_++;
}

void MemBuffer::pk_pack_()
{
    uint8_t *srcbuf = buffer_;
    int srclen;
    int srcoffs;

    //get length of buffer
	srclen=length_;

    /* do the processing */
    uint8_t buf[PK_RUNLEN];
    int len;
    int i;

    srcoffs=0;
    len=0;
    while (1) {
		int st, ed;
		int match_len, match_offs;
		/* find matches */
		st=(srcoffs-PK_WINDOW > 0)?srcoffs-PK_WINDOW:0;
		ed=srcoffs;
		match_offs=0;
		match_len=0;
		for (i=st; i < ed; ++i) {
		    int offs=i;
		    int mlen=0;
		    while (srcbuf[i+mlen] == srcbuf[srcoffs+mlen]) {
				mlen++;
				if (srcoffs+mlen == srclen)
				    break;
				if (mlen == PK_MAXMATCH)
				    break;
			}
			if (mlen > match_len) {
				match_offs=offs;
				match_len=mlen;
	    	}
		}

		/* determine if the match was good enough */
		if ( match_len > 2 ) {
		    /* flush */
		    if (len>0) {
				pk_put_(len);
				for (i=0; i<len; ++i)
				    pk_put_(buf[i]);
				len=0;
		    }
		    /* write a match code */
		    pk_put_(0x80 | match_len);
		    pk_put_(0x100 + (match_offs-srcoffs));    
		    srcoffs+=match_len; /* one already incremented */
		} else {
		    /* store a byte */
		    int data;
		    data=srcbuf[srcoffs];
		    srcoffs++;	

		    buf[len]=data;	
		    len++;
		    /* flush if max run length */
		    if (len==PK_RUNLEN) {
				pk_put_(len);
				for (i=0; i<len; ++i)
				    pk_put_(buf[i]);
				len=0;
		    }
		}

		/* are we all finished? */
		if (srcoffs>=srclen) {
		    break;
		}
	}
	/* flush */
	if (len>0) {
		pk_put_(len);
	for (i=0; i<len; ++i)	
		pk_put_(buf[i]);
	}
    /* end marker */
	pk_put_(0x00);
}


void MemBuffer::pk_store_()
{
	uint8_t* src = buffer_;
    int srclen = length_; 
	int len;

    while (srclen) {
		if(srclen > 0x7f){
			len = 0x7f;
		}else{
			len = srclen;
		}
		pk_put_(len);
		while(len--){
			pk_put_(*src++);
		}
		srclen-=len;
    }
    /* end marker */
    pk_put_(0x00);
}

/* eof */
